/* 
 * Cryptographic API.
 *
 * MD5 Message Digest Algorithm (RFC1321).
 *
 * Derived from cryptoapi implementation, originally based on the
 * public domain implementation written by Colin Plumb in 1993.
 *
 * Copyright (c) Cryptoapi developers.
 * Copyright (c) 2002 James Morris <jmorris@intercode.com.au>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 */

#include "bpalogin.h"

#if 0 /* __LITTLE_ENDIAN */
#define __le32_to_cpus(x) do {} while (0)
#define __cpu_to_le32s(x) do {} while (0)
#else
static inline void  swab32(unsigned int *addr)
{
	unsigned int __x = *addr;
	
	*addr = ((__x & 0x000000FFUL) << 24) |
			((__x & 0x0000FF00UL) <<  8) | 
			((__x & 0x00FF0000UL) >>  8) |
			((__x & 0xFF000000UL) >> 24);
}

#define __le32_to_cpus(x) swab32((x))
#define __cpu_to_le32s(x) swab32((x))
#endif

#define F1(x, y, z)	(z ^ (x & (y ^ z)))
#define F2(x, y, z)	F1(z, x, y)
#define F3(x, y, z)	(x ^ y ^ z)
#define F4(x, y, z)	(y ^ (x | ~z))

#define MD5STEP(f, w, x, y, z, in, s) \
	(w += f(x, y, z) + in, w = (w<<s | w>>(32-s)) + x)


#define MD5_DIGEST_SIZE			16
#define MD5_HMAC_BLOCK_SIZE	64
#define MD5_BLOCK_WORDS		16
#define MD5_HASH_WORDS		4

struct md5_ctx 
{
	unsigned int hash[MD5_HASH_WORDS];
	unsigned int block[MD5_BLOCK_WORDS];

	unsigned int byte_count; /* 32 bit length is enough for currently using. */
};

static void md5_transform(unsigned int *hash, unsigned int const *in)
{
	unsigned int a, b, c, d;

	a = hash[0];
	b = hash[1];
	c = hash[2];
	d = hash[3];

	MD5STEP(F1, a, b, c, d, in[0] + 0xD76AA478, 7);
	MD5STEP(F1, d, a, b, c, in[1] + 0xE8C7B756, 12);
	MD5STEP(F1, c, d, a, b, in[2] + 0x242070DB, 17);
	MD5STEP(F1, b, c, d, a, in[3] + 0xC1BDCEEE, 22);

	MD5STEP(F1, a, b, c, d, in[4] + 0xF57C0FAF, 7);
	MD5STEP(F1, d, a, b, c, in[5] + 0x4787C62A, 12);
	MD5STEP(F1, c, d, a, b, in[6] + 0xA8304613, 17);
	MD5STEP(F1, b, c, d, a, in[7] + 0xFD469501, 22);

	MD5STEP(F1, a, b, c, d, in[8] + 0x698098D8, 7);
	MD5STEP(F1, d, a, b, c, in[9] + 0x8B44F7AF, 12);
	MD5STEP(F1, c, d, a, b, in[10] + 0xFFFF5BB1, 17);
	MD5STEP(F1, b, c, d, a, in[11] + 0x895CD7BE, 22);

	MD5STEP(F1, a, b, c, d, in[12] + 0x6B901122, 7);
	MD5STEP(F1, d, a, b, c, in[13] + 0xFD987193, 12);
	MD5STEP(F1, c, d, a, b, in[14] + 0xA679438E, 17);
	MD5STEP(F1, b, c, d, a, in[15] + 0x49b40821, 22);

	MD5STEP(F2, a, b, c, d, in[1] + 0xF61E2562, 5);
	MD5STEP(F2, d, a, b, c, in[6] + 0xC040B340, 9);
	MD5STEP(F2, c, d, a, b, in[11] + 0x265E5A51, 14);
	MD5STEP(F2, b, c, d, a, in[0] + 0xE9B6C7AA, 20);

	MD5STEP(F2, a, b, c, d, in[5] + 0xD62F105D, 5);
	MD5STEP(F2, d, a, b, c, in[10] + 0x02441453, 9);
	MD5STEP(F2, c, d, a, b, in[15] + 0xD8A1E681, 14);
	MD5STEP(F2, b, c, d, a, in[4] + 0xE7D3FBC8, 20);

	MD5STEP(F2, a, b, c, d, in[9] + 0x21E1CDE6, 5);
	MD5STEP(F2, d, a, b, c, in[14] + 0xC33707D6, 9);
	MD5STEP(F2, c, d, a, b, in[3] + 0xF4D50D87, 14);
	MD5STEP(F2, b, c, d, a, in[8] + 0x455A14ED, 20);

	MD5STEP(F2, a, b, c, d, in[13] + 0xA9E3E905, 5);
	MD5STEP(F2, d, a, b, c, in[2] + 0xFCEFA3F8, 9);
	MD5STEP(F2, c, d, a, b, in[7] + 0x676F02D9, 14);
	MD5STEP(F2, b, c, d, a, in[12] + 0x8D2A4C8A, 20);

	MD5STEP(F3, a, b, c, d, in[5] + 0xFFFA3942, 4);
	MD5STEP(F3, d, a, b, c, in[8] + 0x8771F681, 11);
	MD5STEP(F3, c, d, a, b, in[11] + 0x6D9D6122, 16);
	MD5STEP(F3, b, c, d, a, in[14] + 0xFDE5380C, 23);

	MD5STEP(F3, a, b, c, d, in[1] + 0xA4BEEA44, 4);
	MD5STEP(F3, d, a, b, c, in[4] + 0x4BDECFA9, 11);
	MD5STEP(F3, c, d, a, b, in[7] + 0xF6BB4B60, 16);
	MD5STEP(F3, b, c, d, a, in[10] + 0xBEBFBC70, 23);

	MD5STEP(F3, a, b, c, d, in[13] + 0x289B7EC6, 4);
	MD5STEP(F3, d, a, b, c, in[0] + 0xEAA127FA, 11);
	MD5STEP(F3, c, d, a, b, in[3] + 0xD4EF3085, 16);
	MD5STEP(F3, b, c, d, a, in[6] + 0x04881D05, 23);

	MD5STEP(F3, a, b, c, d, in[9] + 0xD9D4D039, 4);
	MD5STEP(F3, d, a, b, c, in[12] + 0xE6DB99E5, 11);
	MD5STEP(F3, c, d, a, b, in[15] + 0x1FA27CF8, 16);
	MD5STEP(F3, b, c, d, a, in[2] + 0xC4AC5665, 23);

	MD5STEP(F4, a, b, c, d, in[0] + 0xF4292244, 6);
	MD5STEP(F4, d, a, b, c, in[7] + 0x432AFF97, 10);
	MD5STEP(F4, c, d, a, b, in[14] + 0xAB9423A7, 15);
	MD5STEP(F4, b, c, d, a, in[5] + 0xFC93A039, 21);

	MD5STEP(F4, a, b, c, d, in[12] + 0x655B59C3, 6);
	MD5STEP(F4, d, a, b, c, in[3] + 0x8F0CCC92, 10);
	MD5STEP(F4, c, d, a, b, in[10] + 0xFFEFF47D, 15);
	MD5STEP(F4, b, c, d, a, in[1] + 0x85845DD1, 21);

	MD5STEP(F4, a, b, c, d, in[8] + 0x6FA87E4F, 6);
	MD5STEP(F4, d, a, b, c, in[15] + 0xFE2CE6E0, 10);
	MD5STEP(F4, c, d, a, b, in[6] + 0xA3014314, 15);
	MD5STEP(F4, b, c, d, a, in[13] + 0x4E0811A1, 21);

	MD5STEP(F4, a, b, c, d, in[4] + 0xF7537E82, 6);
	MD5STEP(F4, d, a, b, c, in[11] + 0xBD3AF235, 10);
	MD5STEP(F4, c, d, a, b, in[2] + 0x2AD7D2BB, 15);
	MD5STEP(F4, b, c, d, a, in[9] + 0xEB86D391, 21);

	hash[0] += a;
	hash[1] += b;
	hash[2] += c;
	hash[3] += d;
}

static inline void le32_to_cpu_array(unsigned int *buf, unsigned int words)
{
	while (words--) {
		__le32_to_cpus(buf);
		buf++;
	}
}

static inline void cpu_to_le32_array(unsigned int *buf, unsigned int words)
{
	while (words--) {
		__cpu_to_le32s(buf);
		buf++;
	}
}

static inline void md5_transform_helper(struct md5_ctx *ctx)
{
	le32_to_cpu_array(ctx->block, sizeof(ctx->block) / sizeof(unsigned int));
	md5_transform(ctx->hash, ctx->block);
}

static void md5_init(void *ctx)
{
	struct md5_ctx *mctx = ctx;

	mctx->hash[0] = 0x67452301;
	mctx->hash[1] = 0xEFCDAB89;
	mctx->hash[2] = 0x98BADCFE;
	mctx->hash[3] = 0x10325476;
	mctx->byte_count = 0;
}

static void md5_update(void *ctx, const unsigned char *data, unsigned int len)
{
	struct md5_ctx *mctx = ctx;
	const unsigned int avail = sizeof(mctx->block) - (mctx->byte_count & 0x3F);

	mctx->byte_count += len;

	if (avail > len) {
		memcpy((char *)mctx->block + (sizeof(mctx->block) - avail), data, len);
		return;
	}

	memcpy((char *)mctx->block + (sizeof(mctx->block) - avail), data, avail);

	md5_transform_helper(mctx);
	data += avail;
	len -= avail;

	while (len >= sizeof(mctx->block)) {
		memcpy(mctx->block, data, sizeof(mctx->block));
		md5_transform_helper(mctx);
		data += sizeof(mctx->block);
		len -= sizeof(mctx->block);
	}

	memcpy(mctx->block, data, len);
}

static void md5_final(void *ctx, unsigned char *out)
{
	struct md5_ctx *mctx = ctx;
	const unsigned int offset = mctx->byte_count & 0x3F;
	char *p = (char *)mctx->block + offset;
	int padding = 56 - (offset + 1);

	*p++ = 0x80;
	if (padding < 0) {
		memset(p, 0x00, padding + sizeof(unsigned long long));
		md5_transform_helper(mctx);
		p = (char *)mctx->block;
		padding = 56;
	}

	memset(p, 0, padding);
	mctx->block[14] = mctx->byte_count << 3;
	mctx->block[15] = mctx->byte_count >> 29;
	le32_to_cpu_array(mctx->block, (sizeof(mctx->block) -
	                  sizeof(unsigned long long)) / sizeof(unsigned int));
	md5_transform(mctx->hash, mctx->block);
	cpu_to_le32_array(mctx->hash, sizeof(mctx->hash) /sizeof(unsigned int));
	memcpy(out, mctx->hash, sizeof(mctx->hash));
	memset(mctx, 0, sizeof(*mctx));
}

/*
  *  This functions makes the MD5 based data packet which is used to login,
  *  logout and handle heartbeats.
  */
void make_credentials(unsigned char *credentials, struct session *s, uint16 msg, uint32 extra)
{
	int i;
	uint16 nmsg;
	uint32 nextra; /* network byte order. */
	char buf[256];
	struct md5_ctx ctx;

	i = 0;
	nmsg = htons(msg);
	nextra = htonl(extra);

	memcpy(buf, s->nonce, 16);
	i += 16;

	if (s->hashmethod == T_AUTH_MD5_HASH) {
		md5_init(&ctx);
		md5_update(&ctx, (const unsigned char *)s->password, strlen(s->password));
		md5_final(&ctx, (unsigned char *)&buf[i]);
		i += 16;
	} else {
		memcpy(buf + i, s->password, strlen(s->password));
		i += strlen(s->password);
	}
	memcpy(buf + i, &nextra, sizeof(nextra));
	i += sizeof(nextra);
	memcpy(buf + i, &nmsg, sizeof(nmsg));
	i += sizeof(nmsg);

	md5_init(&ctx);
	md5_update(&ctx, (const unsigned char *)buf, i);
	md5_final(&ctx, credentials);
}

