/*
**	BPALogin - lightweight portable BIDS2 login client
**	Copyright (c) 2001-3 Shane Hyde, and others.
** 
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
** 
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
** 
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
*/ 

#include "bpalogin.h"

/*
  * +---------------+---------------+
  * |          2B Type        |      2B Length        |
  * +-------------------------------+
  * |           (Length - 4) Bytes Data ...          |
  * +-------------------------------+
  */

/*
  *  Login to the Authentication server
  *
  *  Returns - 0 - failed to login for some reason.
  *            1 - Logged in successfully
  */
int bpa_login(struct session *s)
{
	int err;
	int result;
	int authsocket;
	uint16 transtype;
	socklen_t addrsize;
	unsigned char credentials[16];
	struct transaction trans;

	result = 0;
	s->localaddr.sin_port = htons(0);

	authsocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	err = bind(authsocket, (struct sockaddr *)&s->localaddr, sizeof(struct sockaddr_in));
	if (err) {
		DEBUGP("Error binding auth socket: %s\n", strerror(errno));
		goto ret;
	}

	err = connect(authsocket,(struct sockaddr *)&s->authhost, sizeof(struct sockaddr_in));
	if (err) {
		DEBUGP("Cant connect to auth server: %s\n", strerror(errno));
		goto ret;
	}

	addrsize = sizeof(struct sockaddr_in);
	getsockname(authsocket, (struct sockaddr *)&s->localipaddress, &addrsize);

	/* start the negotiation */
	start_transaction(&trans, T_MSG_PROTOCOL_NEG_REQ, s->sessionid);

	add_field_int2(&trans, T_PARAM_CLIENT_VERSION, LOGIN_VERSION * 100);
	add_field_string(&trans, T_PARAM_OS_IDENTITY, s->osname);
	add_field_string(&trans, T_PARAM_OS_VERSION, s->osrelease);
	add_field_int2(&trans, T_PARAM_PROTOCOL_LIST, T_PROTOCOL_CHAL);

	send_transaction(authsocket, &trans);

	transtype = receive_transaction(authsocket, &trans);
	if (transtype != T_MSG_PROTOCOL_NEG_RESP) {
		DEBUGP("T_MSG_PROTOCOL_NEG_RESP - error transaction type (%d)\n", transtype);
		goto ret;
	}

	extract_int2_value(&trans, T_PARAM_STATUS_CODE, &s->retcode);
	extract_string_value(&trans, T_PARAM_LOGIN_SERVER_HOST, s->loginserverhost);
	extract_int2_value(&trans, T_PARAM_PROTOCOL_SELECT, &s->protocol);

	if (s->protocol != T_PROTOCOL_CHAL) {
		DEBUGP("T_MSG_PROTOCOL_NEG_RESP - Unsupported protocol (%d)\n", s->protocol);
		goto ret;
	}

	switch(s->retcode) {
	case T_STATUS_SUCCESS:
	case T_STATUS_LOGIN_SUCCESS_SWVER:
		break;

	case T_STATUS_LOGIN_FAIL_SWVER:
		DEBUGP("T_MSG_PROTOCOL_NEG_RESP - Login failure: software version\n");
		goto ret;
	case T_STATUS_LOGIN_FAIL_INV_PROT:
		DEBUGP("T_MSG_PROTOCOL_NEG_RESP - Login failure: invalid protocol\n");
		goto ret;
	case T_STATUS_LOGIN_UNKNOWN:
		DEBUGP("T_MSG_PROTOCOL_NEG_RESP - Login failure: unknown\n");
		goto ret;
	}

	close(authsocket);

	/******************* username && password authentication *******************/

	authsocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	err = bind(authsocket, (struct sockaddr *)&s->localaddr, sizeof(struct sockaddr_in));
	if (err) {
		DEBUGP("Error binding auth socket 2: %s\n", strerror(errno));
		goto ret;
	}

	err = connect(authsocket, (struct sockaddr *)&s->authhost, sizeof(struct sockaddr_in));
	if (err) {
		DEBUGP("Error connecting auth socket 2: %s\n", strerror(errno));
		goto ret;
	}

	start_transaction(&trans, T_MSG_LOGIN_REQ, s->sessionid);

	add_field_string(&trans, T_PARAM_USERNAME, s->username);
	add_field_int2(&trans, T_PARAM_CLIENT_VERSION, LOGIN_VERSION * 100);
	add_field_string(&trans, T_PARAM_OS_IDENTITY, s->osname);
	add_field_string(&trans, T_PARAM_OS_VERSION, s->osrelease);
	add_field_int2(&trans, T_PARAM_REASON_CODE, T_LOGIN_REASON_CODE_NORMAL);
	add_field_int2(&trans, T_PARAM_REQ_PORT, s->listenport);

	send_transaction(authsocket, &trans);

	transtype = receive_transaction(authsocket, &trans);
	if (transtype == T_MSG_LOGIN_RESP)
		goto skippo;

	if (transtype != T_MSG_AUTH_RESP) {
		DEBUGP("T_MSG_AUTH_RESP - error transaction type (%d)\n", transtype);
		goto ret;
	}

	if (!extract_int2_value(&trans, T_PARAM_HASH_METHOD, &s->hashmethod)) {
		DEBUGP("T_MSG_AUTH_RESP - no hashmethod provided\n");
		goto ret;
	}
	if (!extract_string_value(&trans, T_PARAM_NONCE, s->nonce))	{
		DEBUGP("T_MSG_AUTH_RESP - no nonce supplied\n");
		goto ret;
	}

	/*===========================================================*/

	start_transaction(&trans, T_MSG_LOGIN_AUTH_REQ, s->sessionid);

	s->timestamp = time(NULL);
	make_credentials(credentials, s, T_MSG_LOGIN_AUTH_REQ, s->timestamp);

	add_field_data(&trans, T_PARAM_AUTH_CREDENTIALS, (char *)credentials, 16);
	add_field_int4(&trans, T_PARAM_TIMESTAMP, s->timestamp);

	send_transaction(authsocket, &trans);

	transtype = receive_transaction(authsocket, &trans);
skippo:
	if (transtype != T_MSG_LOGIN_RESP) {
		DEBUGP("T_MSG_LOGIN_RESP - error transaction type (%d)\n", transtype);
		goto ret;
	}

	extract_int2_value(&trans, T_PARAM_STATUS_CODE, &s->retcode);
	switch(s->retcode) {
	case T_STATUS_SUCCESS:
	case T_STATUS_LOGIN_SUCCESSFUL_SWVER:
	case T_STATUS_LOGIN_SUCCESSFUL_ALREADY_LOGGED_IN:
		break;

	case T_STATUS_USERNAME_NOT_FOUND:
		DEBUGP("T_MSG_LOGIN_RESP - Login failure: username not known\n");
		goto ret;
		
	case T_STATUS_INCORRECT_PASSWORD:
		DEBUGP("T_MSG_LOGIN_RESP - Login failure: incorrect password\n");
		goto ret;

	case T_STATUS_ACCOUNT_DISABLED:
		DEBUGP("T_MSG_LOGIN_RESP - Login failure: Account disabled\n");
		goto ret;
		
	case T_STATUS_LOGIN_RETRY_LIMIT:
	case T_STATUS_USER_DISABLED:
	case T_STATUS_FAIL_USERNAME_VALIDATE:
	case T_STATUS_FAIL_PASSWORD_VALIDATE:
	case T_STATUS_LOGIN_UNKNOWN:
		DEBUGP("T_MSG_LOGIN_RESP - Login failure: other error\n");
		goto ret;
	}

	extract_int2_value(&trans, T_PARAM_LOGOUT_SERVICE_PORT, &s->logoutport);
	extract_int2_value(&trans, T_PARAM_STATUS_SERVICE_PORT, &s->statusport);
	extract_string_value(&trans, T_PARAM_TSMLIST, s->tsmlist);
	extract_string_value(&trans, T_PARAM_RESPONSE_TEXT, s->resptext);

	/**************************************************************************/
	int i,n;
	char *p, t[200];
	p = s->tsmlist;
	s->tsmcount = 0;

	while ((n = strcspn(p, " ,")) != 0) {
		strncpy(t, p, n);
		t[n] = 0;
		p += n + 1;
		strcpy(s->tsmlist_s[s->tsmcount], t);

		if (!is_dotted_ip(s->tsmlist_s[s->tsmcount]))
			strcat(s->tsmlist_s[s->tsmcount], s->authdomain);

		s->tsmcount++;
	}

	for (i = 0; i < s->tsmcount; i++) {
		struct hostent *host;

		host = gethostbyname(s->tsmlist_s[i]);
		if (host != NULL && host->h_addrtype == AF_INET)
			s->tsmlist_in[i].sin_addr.s_addr = *((unsigned int*)(host->h_addr_list[0]));
		else
			s->tsmlist_in[i].sin_addr.s_addr = inet_addr(s->tsmlist_s[i]);
		DEBUGP("Will accept heartbeats from %s = %s\n", s->tsmlist_s[i],
					inet_ntoa(s->tsmlist_in[i].sin_addr));
	}
	
	DEBUGP("Logged on as %s - successful at %d\n", s->username, uptime());
	result = 1;
	s->sequence = 0;
	s->lastheartbeat = uptime();
	s->recenthb = 0;
	show_bpa_status(1);
	set_bpa_mode(s, 0, 1);

	/*************************************************************************/
	int fd;

	fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (fd >= 0) {
		struct ifreq ifr;

		memset(&ifr, 0, sizeof(struct ifreq));
		ifr.ifr_addr.sa_family = AF_INET;
		strcpy(ifr.ifr_name, s->devname);
		if (ioctl(fd, SIOCGIFADDR, &ifr) == 0) {
			struct sockaddr_in *our_ip;

			our_ip = (struct sockaddr_in *) &ifr.ifr_addr;
			syslog(6, "[Internet connected] IP address: %s,", inet_ntoa(our_ip->sin_addr));
		}
		close(fd);
	}
	/*************************************************************************/

ret:
	if (authsocket >= 0)
		close(authsocket);
	return result;
}

/*
  *  Logout of the BIDS2 system
  *    
  *  Returns - 0 - Could not connect to logout.
  *            1 - Logout successful.
  */
int bpa_logout(uint16 reason, struct session *s)
{
	int err;
	int result;
	int authsocket;
	uint16 transtype;
	unsigned char credentials[16];
	struct transaction trans;

	result = 0;
	s->localaddr.sin_port = htons(0);
	authsocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	err = bind(authsocket, (struct sockaddr *)&s->localaddr, sizeof(struct sockaddr_in));
	if (err) {
		DEBUGP("Error binding logout auth socket: %s\n", strerror(errno));
		goto ret;
	}

	err = connect(authsocket, (struct sockaddr *)&s->authhost, sizeof(struct sockaddr_in));
	if (err) {
		DEBUGP("Error connecting logout auth socket: %s\n", strerror(errno));
		goto ret;
	}

	/* start the negotiation */
	start_transaction(&trans, T_MSG_LOGOUT_REQ, s->sessionid);

	add_field_string(&trans, T_PARAM_USERNAME, s->username);
	add_field_int2(&trans, T_PARAM_CLIENT_VERSION, LOGIN_VERSION * 100);
	add_field_string(&trans, T_PARAM_OS_IDENTITY, s->osname);
	add_field_string(&trans, T_PARAM_OS_VERSION, s->osrelease);
	add_field_int2(&trans, T_PARAM_REASON_CODE, reason);

	send_transaction(authsocket, &trans);

	transtype = receive_transaction(authsocket, &trans);
	if (transtype != T_MSG_AUTH_RESP) {
		DEBUGP("logic error: transtype != T_MSG_AUTH_RESP\n");
		goto ret;
	}

	if (!extract_int2_value(&trans, T_PARAM_HASH_METHOD, &s->hashmethod)) {
		DEBUGP("AUTH: no hashmethod\n");
		goto ret;
	}
	if (!extract_string_value(&trans, T_PARAM_NONCE, s->nonce)) {
		DEBUGP("Auth: no nonce\n");
		goto ret;
	}

	/********************************************************/
	start_transaction(&trans, T_MSG_LOGOUT_AUTH_RESP, s->sessionid);

	s->timestamp = time(NULL);
	make_credentials(credentials, s, T_MSG_LOGOUT_AUTH_RESP, s->timestamp);

	add_field_data(&trans, T_PARAM_AUTH_CREDENTIALS, (char *)credentials, 16);
	add_field_int4(&trans, T_PARAM_TIMESTAMP, s->timestamp);

	send_transaction(authsocket, &trans);

	transtype = receive_transaction(authsocket, &trans);
	if (transtype != T_MSG_LOGOUT_RESP) {
		DEBUGP("logic error: transtype != T_MSG_LOGOUT_RESP\n");
		goto ret;
	}

	extract_int2_value(&trans, T_PARAM_STATUS_CODE, &s->retcode);
	switch(s->retcode) {
	case T_STATUS_SUCCESS:
	case T_STATUS_LOGOUT_SUCCESSFUL_ALREADY_DISCONNECTED:
		break;

	case T_STATUS_USERNAME_NOT_FOUND:
		DEBUGP("Login failure: username not known\n");
		goto ret;

	case T_STATUS_INCORRECT_PASSWORD:
		DEBUGP("Login failure: incorrect password\n");
		goto ret;

	case T_STATUS_ACCOUNT_DISABLED:
		DEBUGP("Login failure: disabled\n");
		goto ret;

	case T_STATUS_LOGIN_RETRY_LIMIT:
	case T_STATUS_USER_DISABLED:
	case T_STATUS_FAIL_USERNAME_VALIDATE:
	case T_STATUS_FAIL_PASSWORD_VALIDATE:
	case T_STATUS_LOGIN_UNKNOWN:
		DEBUGP("Login failure: other error\n");
		goto ret;
	}

	extract_int2_value(&trans, T_PARAM_LOGOUT_SERVICE_PORT, &s->logoutport);
	extract_int2_value(&trans, T_PARAM_STATUS_SERVICE_PORT, &s->statusport);
	extract_string_value(&trans, T_PARAM_TSMLIST, s->tsmlist);
	extract_string_value(&trans, T_PARAM_RESPONSE_TEXT, s->resptext);

	DEBUGP("Logged out successful at %d\n", uptime());
	result = 1;

ret:
	show_bpa_status(0);
	set_bpa_mode(s, 0, 0);

	if (authsocket >= 0)
		close(authsocket);
	return result;
}

/*
  *  Handle heartbeats, wait for the following events to happen -
  *    
  *    1. A heartbeat packet arrives, in which case we reply correctly
  *    2. A timeout occured (ie no heartbeat arrived within 7 minutes)
  *    3. The socket was closed.
  *
  *  Returns - 0 - Heartbeat timeout, and subsequent login failed to connect
  *            1 - Socket closed on us, presuming the user wants to shut down.
  */
int handle_heartbeats(struct session *s)
{
	uint16 transtype;
	struct transaction trans;

	for (;;) {
		transtype = receive_udp_transaction(s, s->listensock, &trans, &s->fromaddr);

		if (transtype == T_STATUS_TIMEOUT) {
			DEBUGP("Timed out waiting for heartbeat - logging on\n");
			/* 
			  * The session needs to be restarted if detected by not seeing incoming heartbeats
			  * for a timeout period of 15 minutes then client should restart the authentication
			  * process.
			  */
			return 0;
		} else if (transtype == T_STATUS_IGNORE_ERR) {
			DEBUGP("Badly structured packet received - discarding\n");
		} else if (transtype == T_STATUS_ERROR) {
			DEBUGP("Socket closed - shutting down\n");
			return 1;
		} else if (transtype == T_MSG_STATUS_REQ) {
			unsigned char buf[16];

			start_transaction(&trans, T_MSG_STATUS_RESP, s->sessionid);
			add_field_int2(&trans, T_PARAM_STATUS_CODE, T_STATUS_TRANSACTION_OK);

			s->sequence++;
			s->hashmethod = T_AUTH_NOHASH;

			make_credentials(buf, s, T_MSG_STATUS_RESP, s->sequence);
			add_field_data(&trans, T_PARAM_STATUS_AUTH, (char *)buf, 16);
			add_field_int4(&trans, T_PARAM_SEQNUM, s->sequence);

			send_udp_transaction(s, &trans);

			s->lastheartbeat = uptime();

			DEBUGP("Responded to heartbeat at %d\n", s->lastheartbeat);
		} else if (transtype == T_MSG_RESTART_REQ) {
			DEBUGP("Restart request - unimplemented\n");
			return 0;
		} else {
			/*
			  *  Melbourne servers were sending spurious UDP packets after authentication
			  *  This works around it.
			  */
			DEBUGP("Unknown heartbeat message %d ", transtype);
		}
	}

	/* Should never get here */
	return 0;
}

