/* SIP extension for UDP NAT alteration.
 *
 * (C) 2005 by Christian Hentschel <chentschel@arnet.com.ar>
 * based on RR's ip_nat_ftp.c and other modules.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/udp.h>

#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include "ip_conntrack_dnisip.h"

#include <linux/ctype.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Christian Hentschel <chentschel@arnet.com.ar>");
MODULE_DESCRIPTION("DNISIP NAT helper");

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

extern struct dnisip_header_nfo ct_dnisip_hdrs[];

static unsigned int mangle_dnisip_packet(struct sk_buff **pskb,
				      enum ip_conntrack_info ctinfo,
				      struct ip_conntrack *ct,
				      const char **dptr, size_t dlen,
				      char *buffer, int bufflen,
				      struct dnisip_header_nfo *hnfo)
{
	unsigned int matchlen, matchoff;

	if (ct_dnisip_get_info(*dptr, dlen, &matchoff, &matchlen, hnfo) <= 0)
		return 0;

	if (!ip_nat_mangle_udp_packet(pskb, ct, ctinfo,
				      matchoff, matchlen, buffer, bufflen))
		return 0;

	/* We need to reload this. Thanks Patrick. */
	*dptr = (*pskb)->data + (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);

	return 1;
}

static unsigned int record_dnisip_branch(struct sk_buff **pskb,
				      enum ip_conntrack_info ctinfo,
				      struct ip_conntrack *ct,
				      const char **dptr, size_t dlen,
				      struct dnisip_header_nfo *hnfo)
{
	enum ip_conntrack_dir dir = CTINFO2DIR(ctinfo);
	unsigned int matchlen, matchoff;
	const char *limit, *aux;
	char *branch;
	struct dnisip_branch_info *entry;

	aux = *dptr;

	limit = *dptr + (dlen - hnfo->lnlen);
	while (aux <= limit) {
		if ((memcmp(aux, hnfo->lname, hnfo->lnlen) == 0)
		    ||(memcmp(aux, hnfo->sname, hnfo->snlen) == 0))
			break;
		aux++;
	}
	if (limit < aux) {
		DEBUGP("record_sip_branch: Not found '%s' or '%s'\n", hnfo->lname, hnfo->sname);
		return 0;
	}

	limit = *dptr + (dlen - hnfo->ln_strlen);
	while (aux <= limit) {
		if ((memcmp(aux, hnfo->ln_str, hnfo->ln_strlen) == 0))
			break;
		aux++;
	}
	if (limit < aux) {
		DEBUGP("record_sip_branch: Not found '%s'\n", hnfo->ln_str);
		return 0;
	}

	if (dir == IP_CT_DIR_REPLY) {
		aux += hnfo->ln_strlen;
		while (aux <= limit) {
			if ((memcmp(aux, hnfo->ln_str, hnfo->ln_strlen) == 0))
				break;
			aux++;
		}
		if (limit < aux) {
			DEBUGP("record_sip_branch: Not found the second '%s'\n",
					hnfo->ln_str);
			return 0;
		}
	}

	aux += hnfo->ln_strlen;
	limit = *dptr + dlen;
	matchoff = aux - (*dptr);
	matchlen = 0;

	while (aux <= limit && (*aux) != '\r' && (*aux) != '\n') {
		aux++;
		matchlen++;
	}

	branch = kmalloc(matchlen + 1, GFP_ATOMIC);
	if (branch == NULL) {
		DEBUGP("record_sip_branch: malloc err!\n");
		return 0;
	}

	memcpy(branch, *dptr + matchoff, matchlen);
	branch[matchlen] = '\0';

	write_lock_bh(&ip_conntrack_lock);
	if (!ct->dnisip_branch) {
		entry = (struct dnisip_branch_info *)
			kmalloc(sizeof(struct dnisip_branch_info), GFP_ATOMIC);

		entry->branch[dir] = branch;
		entry->branch[!dir] = NULL;
		entry->port = 0;
		ct->dnisip_branch = entry;
	} else {
		if (ct->dnisip_branch->branch[dir])
			kfree(ct->dnisip_branch->branch[dir]);
		ct->dnisip_branch->branch[dir] = branch;
	}
	write_unlock_bh(&ip_conntrack_lock);

	DEBUGP("record sip_branch(len:%u):%s\n", matchlen, branch);
	*dptr = (*pskb)->data + (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);

	return 1;
}

static unsigned int check_dnisip_branch(struct sk_buff **pskb,
				     enum ip_conntrack_info ctinfo,
				     struct ip_conntrack *ct,
				     const char **dptr, size_t dlen,
				     struct dnisip_header_nfo *hnfo)
{
	enum ip_conntrack_dir dir = CTINFO2DIR(ctinfo);
	unsigned int matchlen, matchoff, bufflen;
	const char *limit, *aux, *tmp;
	char buffer[16];

	/* "CSeq: xxx REGISTER\r\n" or "CSeq: xxx OPTIONS\r\n" */
	limit = *dptr + dlen;
	aux = ct_dnisip_search("CSeq:", *dptr, sizeof("CSeq:") - 1, dlen);
	if (!aux)
		return 0;

	aux += (sizeof("CSeq:") - 1);
	while ((isdigit(*aux) || (*aux == ' ')) && (aux <= limit))
		aux++;

	*dptr = aux;
	tmp = ct_dnisip_search("\r\n", *dptr, sizeof("\r\n") - 1, limit - aux);
	if (!tmp)
		return 0;

	matchlen = tmp - aux;
	if (((matchlen == 8) && (memcmp(aux, "REGISTER", 8) == 0))
	    || ((matchlen == 7) && (memcmp(aux, "OPTIONS", 7) == 0)))
		goto end;

	DEBUGP("check_sip_branch: there is no 'REGISTER' and 'OPTIONS' in 'CSeq:'\n");
	*dptr = (*pskb)->data + (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);
	
	aux = *dptr;

	limit = *dptr + (dlen - hnfo->lnlen);
	while (aux <= limit) {
		if ((memcmp(aux, hnfo->lname, hnfo->lnlen) == 0)
		    || (memcmp(aux, hnfo->sname, hnfo->snlen) == 0))
			break;
		aux++;
	}
	if (limit < aux) {
		DEBUGP("check_sip_branch: Not found '%s'\n", hnfo->lname);
		return 0;
	}

	limit = *dptr + (dlen - hnfo->ln_strlen);
	while (aux <= limit) {
		if ((memcmp(aux, hnfo->ln_str, hnfo->ln_strlen) == 0))
			break;
		aux++;
	}
	if (limit < aux) {
		DEBUGP("check_sip_branch: Not found '%s'\n", hnfo->ln_str);
		return 0;
	}

	if (dir == IP_CT_DIR_ORIGINAL) {
		aux += hnfo->ln_strlen;
		while (aux <= limit) {
			if ((memcmp(aux, hnfo->ln_str, hnfo->ln_strlen) == 0))
				break;
			aux++;
		}
		if (limit < aux) {
			DEBUGP("check_sip_branch: Not found the second '%s'\n",
					hnfo->ln_str);
			return 0;
		}
	}

	limit = *dptr + dlen;
	aux += hnfo->ln_strlen;
	matchoff = aux - (*dptr);
	matchlen = 0;
	*dptr = aux;

	tmp = ct_dnisip_search(";branch=", *dptr, sizeof(";branch=") - 1, dlen - matchoff);
	if (tmp) {
		DEBUGP("check_sip_branch: ';branch=' is found in 'Via'.\n");
		goto end;
	}

	while (aux <= limit && (*aux) != '\r' && (*aux) != '\n') {
		aux++;
		matchlen++;
	}

	if (ct->dnisip_branch) {
		if (!ip_nat_mangle_udp_packet(pskb, ct, ctinfo,	matchoff, matchlen,
					      ct->dnisip_branch->branch[!dir],
					      strlen(ct->dnisip_branch->branch[!dir])))
			return 0;
		DEBUGP("check_sip_branch: add sip_branch(len: %u):%s\n", 
			strlen(ct->dnisip_branch->branch[!dir]), ct->dnisip_branch->branch[!dir]);

		if (dir == IP_CT_DIR_REPLY) {
			*dptr = (*pskb)->data + (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);
			aux = ct_dnisip_search(";rport", *dptr, sizeof(";rport") - 1, dlen);
			*dptr = (*pskb)->data + (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);
			if (aux) {
				matchoff = aux - *dptr;
				tmp = aux;
				aux += (sizeof(";rport") - 1);
				if (*aux == '=') {
					do {
						aux++;
					} while (aux <= limit && isdigit(*aux));
				}

				matchlen = aux -tmp;
			} else {
				DEBUGP("';rport' not found.\n");
				goto end;
			}

			bufflen = sprintf(buffer, ";rport=%u",
					  ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port));

			if (!ip_nat_mangle_udp_packet(pskb, ct, ctinfo, matchoff,
						      matchlen, buffer, bufflen))
				return 0;

			DEBUGP("check_sip_branch: change rport to '%s'\n", buffer);
		}
	}
end:
	*dptr = (*pskb)->data + (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);
	return 1;
}

static unsigned int ip_nat_dnisip(struct sk_buff **pskb,
			       enum ip_conntrack_info ctinfo,
			       struct ip_conntrack *ct,
			       const char **dptr)
{
	enum ip_conntrack_dir dir = CTINFO2DIR(ctinfo);
	char buffer[sizeof("nnn.nnn.nnn.nnn:nnnnn")];
	unsigned int bufflen, dataoff;
	u_int32_t ip;
	u_int16_t port;

	dataoff = (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);

	/* short packet ? */
	if (((*pskb)->len - dataoff) < (sizeof("SIP/2.0") - 1))
		return 0;

	if (dir == IP_CT_DIR_ORIGINAL) { 
		ip = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
		port = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u.udp.port;
	} else {
		ip = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
		port = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port;
	}
	bufflen = sprintf(buffer, "%u.%u.%u.%u:%u", NIPQUAD(ip), ntohs(port));

#if 0
	if (strnicmp(*dptr, "REGISTER" , 8) == 0)
		DEBUGP("\n*****ip_nat_sip***** REGISTER\n");
	else if (strnicmp(*dptr, "INVITE" , 6) == 0)
		DEBUGP("\n*****ip_nat_sip***** INVITE\n");
	else if (strnicmp(*dptr, "ACK" , 3) == 0)
		DEBUGP("\n*****ip_nat_sip***** ACK\n");
	else if (strnicmp(*dptr, "BYE", 3) == 0)
		DEBUGP("\n*****ip_nat_sip***** BYE\n");
	else if (strnicmp(*dptr, "CANCEL", 6) == 0)
		DEBUGP("\n*****ip_nat_sip***** BYE\n");
	else if (strnicmp(*dptr, "SIP/2.0 200", 11) == 0)
		DEBUGP("\n*****ip_nat_sip***** 200 OK\n");
	else if (strnicmp(*dptr, "SIP/2.0 100", 11) == 0)
		DEBUGP("\n*****ip_nat_sip***** 100 Trying\n");
	else if (strnicmp(*dptr, "SIP/2.0 180", 11) == 0)
		DEBUGP("\n*****ip_nat_sip***** 180 Ringing\n");
	else
		DEBUGP("\n*****ip_nat_sip***** Others\n");

	DEBUGP("dir:%s ctinfo:%d buffer:%s.\n", 
		dir == 0 ? "IP_CT_DIR_ORIGINAL" : "IP_CT_DIR_REPLY", (ctinfo), buffer);
	DEBUGP("ct: %u.%u.%u.%u:%u --> %u.%u.%u.%u:%u\n"
		"\t%u.%u.%u.%u:%u --> %u.%u.%u.%u:%u\n",
		NIPQUAD(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip),
		ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port),
		NIPQUAD(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip),
		ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u.udp.port),
		NIPQUAD(ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip),
		ntohs(ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u.udp.port),
		NIPQUAD(ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip),
		ntohs(ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u.udp.port));
#endif

	/* Status Packet */
	if (memcmp(*dptr, "SIP/2.0", sizeof("SIP/2.0") - 1) == 0) {
		if (!check_dnisip_branch(pskb, ctinfo, ct, dptr, (*pskb)->len - dataoff, &ct_dnisip_hdrs[POS_VIA]))
			return 0;

		if (dir == IP_CT_DIR_REPLY) {
			if (!mangle_dnisip_packet(pskb, ctinfo, ct, dptr, (*pskb)->len - dataoff,
					       buffer, bufflen, &ct_dnisip_hdrs[POS_VIA]))
				return 0;
			DEBUGP("#(Status)# IP_CT_DIR_REPLY Via is changed to %s.\n", buffer);
		}

		mangle_dnisip_packet(pskb, ctinfo, ct, dptr, (*pskb)->len - dataoff,
				  buffer, bufflen, &ct_dnisip_hdrs[POS_CONTACT]);

		return 1;
	}

	/* Request Packet:
	 * REGISTER,OPTIONS (client ---> server)
	 * INVITE,ACK,BYE,CANCEL,SUBSCRIBLE,REFER,NOTIFY,MESSAGE,INFO (client1 -> server -> client2)
	 */
	if (memcmp(*dptr, "REGISTER", sizeof("REGISTER") - 1) != 0
	    && memcmp(*dptr, "OPTIONS", sizeof("OPTIONS") - 1) != 0)
		if (!record_dnisip_branch(pskb, ctinfo, ct, dptr, (*pskb)->len - dataoff,
				       &ct_dnisip_hdrs[POS_VIA]))
			return 0;

	if (dir == IP_CT_DIR_ORIGINAL) { // client ---> server
		if (!mangle_dnisip_packet(pskb, ctinfo, ct, dptr, (*pskb)->len - dataoff,
				       buffer, bufflen, &ct_dnisip_hdrs[POS_VIA]))
			return 0;

		DEBUGP("#(Request)# IP_CT_DIR_ORIGINAL Via is changed to %s.\n", buffer);

		/* Mangle Contact if exists only. */
		mangle_dnisip_packet(pskb, ctinfo, ct, dptr, (*pskb)->len - dataoff,
				  buffer, bufflen, &ct_dnisip_hdrs[POS_CONTACT]);
	} else {
		mangle_dnisip_packet(pskb, ctinfo, ct, dptr, (*pskb)->len - dataoff,
				  buffer, bufflen, &ct_dnisip_hdrs[POS_REQ_HEADER]);
	}

	return 1;
}

static int mangle_content_len(struct sk_buff **pskb,
			      enum ip_conntrack_info ctinfo,
			      struct ip_conntrack *ct,
			      const char *dptr)
{
	unsigned int dataoff, matchoff, matchlen;
	char buffer[sizeof("65536")];
	int bufflen;

	dataoff = (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);

	/* Get actual SDP lenght */
	if (ct_dnisip_get_info(dptr, (*pskb)->len - dataoff, &matchoff,
	                    &matchlen, &ct_dnisip_hdrs[POS_SDP_HEADER]) > 0) {

		/* since ct_dnisip_get_info() give us a pointer passing 'v='
		   we need to add 2 bytes in this count. */
		int c_len = (*pskb)->len - dataoff - matchoff + 2;

		/* Now, update SDP lenght */
		if (ct_dnisip_get_info(dptr, (*pskb)->len - dataoff, &matchoff,
		                    &matchlen, &ct_dnisip_hdrs[POS_CONTENT]) > 0) {

			bufflen = sprintf(buffer, "%u", c_len);

			return ip_nat_mangle_udp_packet(pskb, ct, ctinfo,
							matchoff, matchlen,
							buffer, bufflen);
		}
	}
	return 0;
}

static unsigned int mangle_sdp(struct sk_buff **pskb,
			       enum ip_conntrack_info ctinfo,
			       struct ip_conntrack *ct,
			       u_int32_t newip, u_int16_t port,
			       const char *dptr)
{
	char buffer[sizeof("nnn.nnn.nnn.nnn")];
	unsigned int dataoff, bufflen;

	dataoff = (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);

	DEBUGP("mangle_sdp newip:%u.%u.%u.%u port:%u\n",
			NIPQUAD(newip), ntohs(port));

	/* Mangle owner and contact info. */
	bufflen = sprintf(buffer, "%u.%u.%u.%u", NIPQUAD(newip));
	if (!mangle_dnisip_packet(pskb, ctinfo, ct, &dptr, (*pskb)->len - dataoff,
			       buffer, bufflen, &ct_dnisip_hdrs[POS_OWNER]))
		return 0;

	if (!mangle_dnisip_packet(pskb, ctinfo, ct, &dptr, (*pskb)->len - dataoff,
			       buffer, bufflen, &ct_dnisip_hdrs[POS_CONNECTION]))
		return 0;

	/* Mangle media port. */
	bufflen = sprintf(buffer, "%u", port);
	if (!mangle_dnisip_packet(pskb, ctinfo, ct, &dptr, (*pskb)->len - dataoff,
			       buffer, bufflen, &ct_dnisip_hdrs[POS_MEDIA]))
		return 0;

	return mangle_content_len(pskb, ctinfo, ct, dptr);
}

static int kill_sdp_nat_ct(struct ip_conntrack *i, void *data)
{
	return ip_ct_tuple_equal(
		(const struct ip_conntrack_tuple *)&i->tuplehash[IP_CT_DIR_ORIGINAL].tuple,
		(const struct ip_conntrack_tuple *)data);
}


void ip_nat_sdp_expect(struct ip_conntrack *ct,
		       struct ip_conntrack_expect *exp)
{
	struct ip_nat_range range;

	/* This must be a fresh one. */
	BUG_ON(ct->status & IPS_NAT_DONE_MASK);

	/* Change src to where master sends to */
	range.flags = IP_NAT_RANGE_MAP_IPS;
	range.min_ip = range.max_ip
		= ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
	/* hook doesn't matter, but it has to do source manip */
	ip_nat_setup_info(ct, &range, NF_IP_POST_ROUTING);

	/* For DST manip, map port here to where it's expected. */
	range.flags = (IP_NAT_RANGE_MAP_IPS | IP_NAT_RANGE_PROTO_SPECIFIED);
	range.min = range.max = exp->saved_proto;
	range.min_ip = range.max_ip
		= ct->master->tuplehash[!exp->dir].tuple.src.ip;
	/* hook doesn't matter, but it has to do destination manip */
	ip_nat_setup_info(ct, &range, NF_IP_PRE_ROUTING);

	ip_ct_iterate_cleanup(kill_sdp_nat_ct, &ct->tuplehash[IP_CT_DIR_REPLY].tuple);
}

/* So, this packet has hit the connection tracking matching code.
   Mangle it, and change the expectation to match the new version. */
static unsigned int ip_nat_sdp(struct sk_buff **pskb,
			       enum ip_conntrack_info ctinfo,
			       struct ip_conntrack *ct,
			       struct ip_conntrack_expect *exp,
			       const char *dptr)
{
	enum ip_conntrack_dir dir = CTINFO2DIR(ctinfo);
	u_int32_t newip;
	u_int16_t port = 0;

	DEBUGP("#####ip_nat_sdp##### ctinfo:%u\n", ctinfo);

	if (dir == IP_CT_DIR_REPLY) {
		unsigned int dataoff, datalen;
		int matchoff, matchlen;

		dataoff = (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);
		datalen = (*pskb)->len - dataoff;

		if (ct_dnisip_get_info(dptr, datalen, &matchoff, &matchlen,
				    &ct_dnisip_hdrs[POS_CONNECTION]) <= 0)
			return NF_ACCEPT;

		if (parse_ipaddr(dptr + matchoff, NULL, &newip, dptr + datalen) < 0)
			return NF_DROP;

		if (ct_dnisip_get_info(dptr, datalen, &matchoff, &matchlen,
				    &ct_dnisip_hdrs[POS_MEDIA]) <= 0)
			return NF_ACCEPT;

		port = simple_strtoul(dptr + matchoff, NULL, 10);
		if (port < 1024)
			return NF_DROP;

		struct ip_conntrack_expect *e;
		struct ip_conntrack_tuple tuple =
				{ { ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip, { 0 } },
				  { newip, { htons(port) }, IPPROTO_UDP } };

		DEBUGP("IP_CT_DIR_REPLY tuple %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u\n",
			NIPQUAD(tuple.src.ip), ntohs(tuple.src.u.udp.port),
			NIPQUAD(tuple.dst.ip), ntohs(tuple.dst.u.udp.port));

		e =  ip_conntrack_expect_find(&tuple);
		if (e) {
			port = ntohs(e->saved_proto.udp.port);
			newip = e->master->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;

			if (!mangle_sdp(pskb, ctinfo, ct, newip, port, dptr)) {
				ip_conntrack_unexpect_related(e);
				return NF_DROP;
			}

			ip_conntrack_unexpect_related(e);
		}

		return NF_ACCEPT;
	}

	if (!exp) {
		DEBUGP("Error, there is no exp!\n");
		return NF_DROP;
	}

	/* Connection will come from reply */
	newip = ct->tuplehash[!dir].tuple.dst.ip;

	exp->tuple.dst.ip = newip;
	exp->saved_proto.udp.port = exp->tuple.dst.u.udp.port;
	exp->dir = !dir;

	/* When you see the packet, we need to NAT it the same as the
	   this one. */
	exp->expectfn = ip_nat_sdp_expect;

	DEBUGP("exp tuple: "); DUMP_TUPLE(&exp->tuple);
	DEBUGP("mask:  "); DUMP_TUPLE(&exp->mask);

	/* Try to get same port: if not, try to change it. */
	for (port = ntohs(exp->saved_proto.udp.port); port != 0; port += 2) {
		exp->tuple.dst.u.udp.port = htons(port);
		if (ip_conntrack_expect_related(exp) == 0)
			break;
	}

	if (port == 0)
		return NF_DROP;

	if (!mangle_sdp(pskb, ctinfo, ct, newip, port, dptr)) {
		ip_conntrack_unexpect_related(exp);
		return NF_DROP;
	}

	return NF_ACCEPT;
}

static void __exit fini(void)
{
	ip_nat_dnisip_hook = NULL;
	ip_nat_sdp_hook = NULL;
	/* Make sure noone calls it, meanwhile. */
	synchronize_net();
}

static int __init init(void)
{
	BUG_ON(ip_nat_dnisip_hook);
	BUG_ON(ip_nat_sdp_hook);
	ip_nat_dnisip_hook = ip_nat_dnisip;
	ip_nat_sdp_hook = ip_nat_sdp;
	return 0;
}

module_init(init);
module_exit(fini);
